/*
 * Copyright (c) 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <DemUnidraw/demcmds.h>
#include <DemUnidraw/demcomps.h>
#include <DemUnidraw/demeditor.h>
#include <DemUnidraw/demfunc.h>
#include <DemUnidraw/demviewer.h>
#include <DemVtk/demtriangle.h>
#include <Unidraw/clipboard.h>
#include <ComTerp/comhandler.h>
#include <ComTerp/comterpserv.h>
#include <ComTerp/comvalue.h>
#include <stream.h>

extern DemTriangle* cell(float px, float py);

extern float /*DemCameraCmd::*/elev(float px, float py);

/*****************************************************************************/

ElevFunc::ElevFunc(ComTerp* comterp) : ComFunc(comterp)
{}

void ElevFunc::execute() {
    ComValue utme = stack_arg(0);
    ComValue utmn = stack_arg(1);
    reset_stack();
    ComValue result;

    ComTerpServ* terp = (ComTerpServ*)comterp();
    ComterpHandler* handler = (ComterpHandler*)terp->handler();
    if (handler) {
      float x, y, z;
      x = utme.float_val();
      y = utmn.float_val();
      z = elev(x, y);
      filebuf obuf(handler->get_handle());
      ostream ostr(&obuf);
      ostr.form("%f\n", z);
      ostr.flush();
    }
#if 0
    result.float_ref() = z;
    result.type(AttributeValue::FloatType);
    push_stack(result);
#endif
    return;
}

/*****************************************************************************/

CellFunc::CellFunc(ComTerp* comterp) : ComFunc(comterp)
{}

void CellFunc::execute() {
    ComValue utme(stack_arg(0));
    ComValue utmn(stack_arg(1));
    reset_stack();
    ComValue result;

    ComTerpServ* terp = (ComTerpServ*)comterp();
    ComterpHandler* handler = (ComterpHandler*)terp->handler();
    if (handler) {
      float x, y, z;
      x = utme.float_val();
      y = utmn.float_val();
      DemTriangle* tr = cell(x, y);
      float *pt1, *pt2, *pt3;
      pt1 = tr->Points.GetPoint(1);
      pt2 = tr->Points.GetPoint(2);
      pt3 = tr->Points.GetPoint(0);
      filebuf obuf(handler->get_handle());
      ostream ostr(&obuf);
      ostr.form("%f %f %f %f %f %f %f %f %f\n",
		pt1[0], pt1[1], pt1[2],
		pt2[0], pt2[1], pt2[2],
		pt3[0], pt3[1], pt3[2]);
      ostr.flush();
    }
    return;
}

/*****************************************************************************/

VehicleFunc::VehicleFunc(ComTerp* comterp, Editor* ed) : UnidrawFunc(comterp, ed)
{
  _name_symval = symbol_add("name");
}

void VehicleFunc::execute() {
    ComValue& namestr = stack_key(symbol_find("name"));
    
    cerr << "name of vehicle is " << namestr.string_ptr() << "\n";
    reset_stack();
    // paste a new actorcomp for vehicle
    ActorComp* acomp = new ActorComp(namestr.string_ptr());
    PasteCmd* paste = new PasteCmd(_ed, new Clipboard(acomp));
    execute_log(paste);
}

/*****************************************************************************/

MoveToFunc::MoveToFunc(ComTerp* comterp, Editor* ed) 
: UnidrawFunc(comterp, ed)
{
  _xdir_symval = symbol_add("xdir");
  _ydir_symval = symbol_add("ydir");
  _heading_symval = symbol_add("heading");
}

void MoveToFunc::execute() {
    ComValue compval = stack_arg(0, true);
    ComValue xloc = stack_arg(1);
    ComValue yloc = stack_arg(2);
    ComValue xdir = stack_key(_xdir_symval);
    ComValue ydir = stack_key(_ydir_symval);
    ComValue heading = stack_key(_heading_symval);
    reset_stack();
#if 0
    if (strcmp(((DemEditor*)_ed)->VehicleName(),
	       compval.symbol_ptr())==0) 
      {
	DemViewer* viewer = (DemViewer*)_ed->GetViewer();
	float dirx, diry, thisx, thisy;
	thisx = xloc.float_val();
	thisy = yloc.float_val();
	dirx = xdir.float_val();
	diry = ydir.float_val();
	CameraCmd* cmd = new CameraCmd(thisx, thisy, 
				       thisx + 100*dirx, thisy + 100*diry, false);
	cmd->SetEditor(_ed);
	execute_log(cmd);
	push_stack(ComValue::nullval());
	return;
      }
    else { // update some other vehicle in the world

    }
#endif
    // move actorcomp cmd
    DemMoveActorCmd* movecmd = new DemMoveActorCmd(_ed, (char*)compval.symbol_ptr(),
						   xloc.float_val(),
						   yloc.float_val(),
						   xdir.float_val(),
						   ydir.float_val());
    execute_log(movecmd);
}
