/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Map/mapdb.h>
#include <Map/maplib.h>

#include <iostream.h>
#include <string.h>

/*****************************************************************************/

MapDatabase::MapDatabase (const char* name, const char* path) : MapFiles(name, path) {
    _nlibs = -1;
    _libraries = nil;
}

MapDatabase::~MapDatabase () {
    for (int i=0; i<_nlibs; i++) 
	delete _libraries[i];
    delete [] _libraries;
}

int MapDatabase::nlibs() const{
    if (!initialized()) ((MapDatabase*)this)->init_libraries();

    return _nlibs;
}

MapLibrary* MapDatabase::library(unsigned libnum) const{
    if (!initialized()) ((MapDatabase*)this)->init_libraries();
    
    if (libnum<_nlibs) 
	return _libraries[libnum];
    else
	return nil;
}
     
MapLibrary* MapDatabase::library(const char* libname) const{
    if (!initialized()) ((MapDatabase*)this)->init_libraries();
    
    for (int i = 0; i < _nlibs; i++) {
	if (strcmp(libname,_libraries[i]->name())==0)
	    return _libraries[i];
    }
    return nil;
}

const char* MapDatabase::description() const { return "generic database description"; }

ostream& operator<< (ostream& out, const MapDatabase& db) {
    out << "Database:  " << db.path() << "\n";
    out << "Description:  " << db.description() << "\n";
    out << "Libraries:  " << db.nlibs() << "\n";
    for (int i = 0; i < db.nlibs(); i++) {
	out << db.library(i);
    }
    return out;
}

void MapDatabase::init_libraries() {
    if (initialized()) return;

    char **libnames = library_names(_nlibs);
    if (_nlibs==0) return;

    _libraries = new MapLibrary*[_nlibs];
    for (int i=0; i<_nlibs; i++) {
	_libraries[i] = make_library(libnames[i]);
	delete libnames[i];
    }
    delete [] libnames;
}

