/*
 * Copyright 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <MapUnidraw/geoloctool.h>
#include <MapUnidraw/mapclasses.h>
#include <MapUnidraw/mapeditor.h>
#include <MapUnidraw/mapviewer.h>
#include <MapUnidraw/mapviewstate.h>

#include <IVGlyph/gdialogs.h>

#include <InterViews/event.h>
#include <InterViews/transformer.h>
#include <InterViews/window.h>

#include <iostream.h>
#include <stdio.h>

/*****************************************************************************/

ClassId GeoLocTool::GetClassId () { return GEOLOC_TOOL; }

boolean GeoLocTool::IsA (ClassId id) {
    return GEOLOC_TOOL == id || Tool::IsA(id);
}

GeoLocTool::GeoLocTool (ControlInfo* m) : Tool(m)
{
}

Tool* GeoLocTool::Copy () { return new GeoLocTool(CopyControlInfo()); }

Manipulator* GeoLocTool::CreateManipulator (
    Viewer* v, Event& e, Transformer* rel
) {
  MapViewer* viewer = (MapViewer*)v;
  float lon, lat;
  viewer->ScreenToGeo(e.x, e.y, lon, lat);
  char buffer[BUFSIZ];
  sprintf( buffer, "longitude,latitude:  %.2f %c  %.2f %c", 
	   fabs(lon), lon>=0.0 ? 'E' : 'W', 
	   fabs(lat), lat>=0.0 ? 'N' : 'S');
  GAcknowledgeDialog* dialog = new GAcknowledgeDialog(buffer);
  Resource::ref(dialog);
  dialog->post_for(v->GetEditor()->GetWindow());
  Resource::unref(dialog);
  Manipulator* m = nil;
  return m;

}
