/*
 * Copyright 1994, 1995 Cartoactive Systems, Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

/*
 * MapViewer - a OverlayViewer with support for map operations
 */

#ifndef mapviewer_h
#define mapviewer_h

#include <FrameUnidraw/frameviewer.h>

#include <IV-2_6/_enter.h>

class OverlayComp;
class OverlayPage;

class MapViewer : public FrameViewer {
public:
    MapViewer(
        Editor*, GraphicView*, UPage*, Grid* = nil, 
        Coord = 0, Coord = 0, Orientation = Normal,
	Alignment = Center, Zooming = Continuous
    );
    virtual void Draw();
    virtual void Redraw(Coord, Coord, Coord, Coord);

    void DrawingToGeo(float dx, float dy, float& lon, float& lat);
    void GeoToDrawing(float lon, float lat, float& dx, float& dy);
    void ScreenToGeo(int sx, int sy, float& lon, float& lat);
    void GeoToScreen(float lon, float lat, int& sx, int& sy);

    void DrawingToUtm(float dx, float dy, float& utm_e, float& utm_n,
		      int& zone);
    void UtmToDrawing(float utm_e, float utm_n, int zone,
		      float& dx, float& dy);
    void ScreenToUtm(int sx, int sy, float& utm_e, float& utm_n);
    void UtmToScreen(float utm_e, float utm_n, int zone, int& sx, int& sy);

    void SetMapPage(OverlayComp*);
    static OverlayPage* make_map_page(OverlayComp*);
protected:
    virtual float LimitMagnification(float);

};

#include <IV-2_6/_leave.h>

#endif
