/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Vpf/vpfcov.h>
#include <Vpf/vpffclass.h>
#include <Vpf/vpftiledfc.h>
#include <Vpf/vpflib.h>

#include <iostream.h>
#include <string.h>
#include <sys/stat.h>

extern "C" {
#include <malloc.h>
void format_date( date_type date, char *fmtdate );
}

/*****************************************************************************/

VpfCoverage::VpfCoverage (VpfLibrary* lib, const char* name)
: MapCoverage(lib, name) {
    _desc = NULL;
    library(lib);
}


VpfCoverage::~VpfCoverage () {
    delete _desc;
    if (_intvdt.fp)
      vpf_close_table(&_intvdt);
    if (_charvdt.fp)
      vpf_close_table(&_charvdt);
    if (_edgetable.fp)
	vpf_close_table(&_edgetable);
    if (_facetable.fp)
	vpf_close_table(&_facetable);
}

void VpfCoverage::library(VpfLibrary* lib) { 
    _library = (MapLibrary*)lib; 
    if (_library) {
        description();
        init_feature_classes();
    }
}

const char* VpfCoverage::path() const {
    if (_library && !_path) {
        MapCoverage::path();
#if 0
        struct stat buf;
        char intpath[256];
	const char* libpath = _library->path();
        strcpy(intpath, libpath);
        strcat(intpath, "/");
	const char* name = _name;
        strcat(intpath, name);
        strcat(intpath, "/int.vdt");
        ((VpfCoverage*)this)->_intvdt.fp = NULL;
        if (stat(intpath, &buf) == 0)
	    ((VpfCoverage*)this)->_intvdt = vpf_open_table(intpath,disk,"rb",NULL);
        char charpath[256];
        strcpy(charpath, libpath);
        strcat(charpath, "/");
        strcat(charpath, name);
        strcat(charpath, "/char.vdt");
        ((VpfCoverage*)this)->_charvdt.fp = NULL;
        if (stat(charpath, &buf) == 0)
	    ((VpfCoverage*)this)->_charvdt = vpf_open_table(charpath,disk,"rb",NULL);
#endif
        ((VpfCoverage*)this)->_edgetable.fp = NULL;
        ((VpfCoverage*)this)->_facetable.fp = NULL;
    }
    return _path;
}

VpfFeatureClass* VpfCoverage::feature_class(int fclsnum) const {
    if (fclsnum >= 0)
        return (VpfFeatureClass*)MapCoverage::feature_class(fclsnum);
    else
       return nil;
}

VpfFeatureClass* VpfCoverage::feature_class(const char* name) const {
    return (VpfFeatureClass*)MapCoverage::feature_class(name);
}

char* VpfCoverage::int_value_description(const char* colname, int val) {
        struct stat buf;
        char intpath[256];
	const char* libpath = _library->path();
        strcpy(intpath, libpath);
        strcat(intpath, "/");
	const char* name = _name;
        strcat(intpath, name);
        strcat(intpath, "/int.vdt");
        ((VpfCoverage*)this)->_intvdt.fp = NULL;
        if (stat(intpath, &buf) == 0)
	    ((VpfCoverage*)this)->_intvdt = vpf_open_table(intpath,disk,"rb",NULL);

    long int attpos = table_pos("ATTRIBUTE", _intvdt);
    long int valpos = table_pos("VALUE", _intvdt);
    long int dscpos = table_pos("DESCRIPTION", _intvdt);
    row_type row;
    char* col;
    char* retval = NULL;
    long int n;
    int v;
    for (int i = 1; i <= _intvdt.nrows; i++) {
	row = get_row(i, _intvdt);
	col = (char *)get_table_element(attpos, row, _intvdt, NULL, &n);
	rightjust(col);
	get_table_element(valpos, row, _intvdt, &v, &n);
	if (strcmp(col, colname)==0 && val==v) {
	    retval = (char *)get_table_element(dscpos, row, _intvdt, NULL, &n);
	    free(col);
	    vpf_close_table(&_intvdt);
	    return retval;
	}
	free(col);
    }
    vpf_close_table(&_intvdt);
    return retval;
}

char* VpfCoverage::char_value_description(const char* colname, char* val) {
        struct stat buf;
        char charpath[256];
	const char* libpath = _library->path();
        strcpy(charpath, libpath);
        strcat(charpath, "/");
	const char* name = _name;
        strcat(charpath, name);
        strcat(charpath, "/char.vdt");
        ((VpfCoverage*)this)->_charvdt.fp = NULL;
        if (stat(charpath, &buf) == 0)
	    ((VpfCoverage*)this)->_charvdt = vpf_open_table(charpath,disk,"rb",NULL);

    long int attpos = table_pos("ATTRIBUTE", _charvdt);
    long int valpos = table_pos("VALUE", _charvdt);
    long int dscpos = table_pos("DESCRIPTION", _charvdt);
    row_type row;
    char* col;
    char* retval = NULL;
    long int n;
    char* v;
    for (int i = 1; i <= _charvdt.nrows; i++) {
	row = get_row(i, _charvdt);
	col = (char *)get_table_element(attpos, row, _charvdt, NULL, &n);
	rightjust(col);
	v = (char*)get_table_element(valpos, row, _charvdt, NULL, &n);
	if (strcmp(col, colname)==0 && strcmp(v,val)==0) {
	    retval = (char *)get_table_element(dscpos, row, _charvdt, NULL, &n);
	    free(col);
	    vpf_close_table(&_charvdt);
	    return retval;
	}
	free(col);
    }
    vpf_close_table(&_charvdt);
    return retval;
}

const char* VpfCoverage::description() const {
    if (!_desc) {
	vpf_table_type table;
	row_type row;
	char catpath[255];
	strcpy(catpath, _library->path());
	strcat(catpath, "/cat.");
	table = vpf_open_table(catpath, disk, "rb", NULL);
	if (!table.fp) {
	    return NULL;
	}
	else {
	    long int covfieldnum = table_pos("COVERAGE_NAME", table);
	    if (covfieldnum < 0) {
		cerr << "VpfCoverage::description: ";
		cerr << "Invalid CAT (" << _library->path() << ") - missing COVERAGE_NAME field\n";
		vpf_close_table(&table);
		return NULL;
	    }
	    else {
		long int descfieldnum = table_pos("DESCRIPTION",table);
		if (descfieldnum < 0) {
		    cerr << "VpfCoverage::description: ";
		    cerr << "Invalid CAT (" << _library->path() << ") - missing DESCRIPTION field\n";
		    vpf_close_table(&table);
		    return NULL;
		}
		else {
		    long int n;
		    char* cov;
		    int found = 0;
		    for (int i = 0; i < table.nrows+1; i++) {
			if (i < table.nrows) {
			    row = read_next_row(table);
			    cov = (char *)get_table_element(covfieldnum,row,table,NULL,&n);
			    rightjust(cov);
			    if (strcasecmp(cov, _name) == 0) {
				found = 1;
				((VpfCoverage*)this)->_desc = (char *)get_table_element(descfieldnum,row,table,NULL,&n);
			    }
			    free(cov);
			    free_row(row,table);
			    if (found) break;
			}
			else if (strcasecmp("tileref", _name)==0) {
			    ((VpfCoverage*)this)->_desc = strdup("Tile Reference");
			    found = 1;
			    break;
			}
			else if (strcasecmp("libref", _name)==0) {
			    ((VpfCoverage*)this)->_desc = strdup("Library Reference");
			    found = 1;
			    break;
			}
		    }
		    //vpf_close_table(&table);
		    if (!found) {
			cerr << "VpfCoverage_description: Coverage " << _name << 
			    " not found for library " << _library->path() << "\n";
		    }
		}
	    }
	    vpf_close_table(&table);
	}
	rightjust(((VpfCoverage*)this)->_desc);
    }
    return _desc;
}

char** VpfCoverage::feature_class_names (int& nfcls) const {
    char** fnames = (char **)malloc(sizeof(char *));
    struct stat buf;
    long int i,j,k,n;
    char** ptr;
    vpf_table_type table;
    row_type row;
    char fcspath[255], fcspath2[255];
    strcpy(fcspath, _library->path());
    strcat(fcspath, "/");
    strcat(fcspath, _name);
    strcpy(fcspath2, fcspath);
    strcat(fcspath, "/fcs.");
    strcat(fcspath2, "/fcs");
    if (stat(fcspath, &buf) == 0)
	table = vpf_open_table(fcspath, disk, "rb", NULL);
    else if (stat(fcspath2, &buf) == 0)
	table = vpf_open_table(fcspath2, disk, "rb", NULL);
    else
        table.fp = NULL;
    if (!table.fp) {
	if (strcmp(this->name(), "libref")==0) {
	    nfcls = 1;
	    fnames[0] = strdup("libref");
	}
	else {
#if 0
	    cerr << "VpfCoverage::feature_class_names: Can't open "
		 << fcspath << "\n";
#endif
	    nfcls = 0;
	    fnames = NULL;
	}
    }
    else {
	long int fcfieldnum = table_pos("FEATURE_CLASS", table);
	if (fcfieldnum < 0) {
	    cerr << "VpfCoverage::feature_class_names: ";
	    cerr << "Invalid FCS (" << fcspath << ") - missing FEATURE_CLASS field\n";
	    vpf_close_table(&table);
	    nfcls = 0;
	    fnames = NULL;
	}
	else {
	    int found;
	    char* fc;
	    j = 0;
	    row = read_next_row(table);
	    fnames[j] = (char *)get_table_element(fcfieldnum,row,table,NULL,&n);
	    rightjust(fnames[j]);
	    strlower(fnames[j]);
	    free_row(row,table);
	    
	    for (i = 2; i <= table.nrows; i++) {
		row = read_next_row(table);
		fc = (char *)get_table_element(fcfieldnum,row,table,NULL,&n);
		rightjust(fc);
		free_row(row,table);
		found = 0;
		for (k=j; k >= 0; k--) {
		    if (strcasecmp(fc,fnames[k])==0) {
			found = 1;
			break;
		    }
		}
		if (!found) {
		    j++;
		    ptr = (char **)realloc(fnames,(j+1)*sizeof(char *));
		    if (!ptr) {
			cerr << "VpfCoverage::feature_class_names: ";
			cerr << "Memory allocation error\n";
			for (k=(j-1);k>=0;k--) {
			    free(fnames[k]);
			}
			free(fnames);
			vpf_close_table(&table);
			return (char **)NULL;
		    }
		    fnames = ptr;
		    fnames[j] = (char *)malloc(strlen(fc)+1);
		    if (!fnames[j]) {
			cerr << "VpfCoverage::feature_class_names: ";
			cerr << "Memory allocation error\n";
			for (k=0;k<j;k++) free(fnames[k]);
			free(fnames);
			vpf_close_table(&table);
			return (char **)NULL;
		    }
		    strcpy(fnames[j],fc);
		    strlower(fnames[j]);
		}
	    }
	    vpf_close_table(&table);
	    nfcls = j+1;
	}
    }
    return fnames;
}

void VpfCoverage::print() const {
    printf("  Name: %s\n", _name);
    printf("  Description: %s\n", _desc);
    printf("  FCS table dump:\n");
    dump_fcs();
    printf("  Feature Classes: %d\n", _nfcls);
    for (int i = 0; i < _nfcls; i++)
	feature_class(i)->print();
}

void VpfCoverage::dump_fcs() const {
    char path[256];
    strcpy(path, _library->path());
    strcat(path, "/");
    strcat(path, _name);
    strcat(path, "/");
    strcat(path, "fcs.");
    int val = pretty_dump_table(path, stdout);
    printf("return value: %d\n", val);
}

MapFeatureClass* VpfCoverage::make_feature_class(const char* name) { 
    if (library()->tiled() && strcmp(name, "tileref")!=0 &&
	strcmp(name, "tilereft")!=0 &&
	strcmp(name, "libref")!=0)
	return new VpfTiledFeatureClass(this,name);
    else
	return new VpfFeatureClass(this,name);
}
					

void VpfCoverage::load_edge_table(char* tn) {
    char edgepath[255];
    strcpy(edgepath, path());
    if (tn) {
      strcat(edgepath, "/");
      strcat(edgepath, tn);
    }
    strcat(edgepath, "/edg.");
    _edgetable = vpf_open_table(edgepath,ram,"rb",NULL);
}
void VpfCoverage::load_face_table(char* tn) {
    char facepath[255];
    strcpy(facepath, path());
    if (tn) {
      strcat(facepath, "/");
      strcat(facepath, tn);
    }
    strcat(facepath, "/fac.");
    _facetable = vpf_open_table(facepath,ram,"rb",NULL);
}
void VpfCoverage::load_node_table() {}

vpf_table_type VpfCoverage::edgetable(char* tn) {
    if (_edgetable.fp == NULL)
	load_edge_table(tn);
    return _edgetable;
}

vpf_table_type VpfCoverage::facetable(char* tn) {
    if (_facetable.fp == NULL)
	load_face_table(tn);
    return _facetable;
}


/*************************************************************************
*
*N  pretty_dump_table
*
*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
*
*   Purpose:
*P
*     This function pretty dumps the contents of a VPF table to a text file.
*E
*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
*
*   Parameters:
*A
*     tablepath      <input>  == (char *) vpf table to dump.
*     outfp          <input>  == (FILE *) pointer to ASCII file to dump to.
*     vpf_dump_table <output> == (int) the status of the vpf_dump_table 
*				 function.
*				 0 - success
*                                1 = table not found
*                                2 = access violation for tablepath
*                                3 = invalid VPF table
*E
*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
*
*   History:
*H
*    Barry Michaels                            DOS Turbo C
*    Scott Simon	Oct 1992               UNIX mdb version, modified
*					       DOS Turbo C vpf_dump_table
*    Jorge Gautier      Nov 1993               Pretty version
*E
*************************************************************************/
int pretty_dump_table( char *tablepath,
                    FILE *outfp ) {
   vpf_table_type table;
   long int   i,j,k,n;
   short int  ival,*iptr;
   long int   lval,*lptr;
   float      fval,*fptr;
   double     rval,*rptr;
   date_type  dval,*dptr;
   id_triplet_type kval, *kptr;
   coordinate_type cval, *cptr;
   double_coordinate_type bval, *bptr;
   tri_coordinate_type zval, *zptr;
   double_tri_coordinate_type yval, *yptr;
   char       *buf,  ch, date[40];
   row_type   row;

   if (!is_vpf_table(tablepath)) return 3;

   table = vpf_open_table(tablepath,disk,"rb",NULL);
   if (!table.fp) return 3;

   fprintf(outfp,"%s\n%s\n\n",tablepath,table.description);

   fprintf(outfp,"Definition:\n");
   for (i=0;i<table.nfields;i++) {
      if (table.header[i].count < 0)
	 fprintf(outfp,"%s (%c,*)  %s\n",
		 table.header[i].name,table.header[i].type,
		 table.header[i].description);
      else
	 fprintf(outfp,"%s (%c,%ld)  %s\n",
		 table.header[i].name,table.header[i].type,
		 table.header[i].count,table.header[i].description);
   }

   fprintf(outfp,"\nContents:\n");

   for (i=0;i<table.nfields;i++) {
      if (table.header[i].count < 0)
	 fprintf(outfp,"%s ",
		 table.header[i].name);
      else
	 fprintf(outfp,"%s ",
		 table.header[i].name);
   }
   fprintf(outfp, "\n");

   for (i=1;i<=table.nrows;i++) {
      row = read_next_row(table);
      for (j=0;j<table.nfields;j++) {
	 switch (table.header[j].type) {
	    case 'T':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&ch,&n);
		  fprintf(outfp,"%c ",ch);
	       } else {
		  buf = (char *)get_table_element(j,row,table,NULL,&n);
		  n = strlen(table.header[j].name) + 2;
		  for (k=0;k<strlen(buf);k++) {
		     fprintf(outfp,"%c",buf[k]);
		     n++;
		     if (n>80) {
			n = 0;
		     }
		  }
		  fprintf(outfp, " ");
		  if (buf) free(buf);
	       }
	       break;
	    case 'I':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&lval,&n);
		  if (lval != NULLINT)
		     fprintf(outfp,"%ld ",lval);
		  else
		     fprintf(outfp,"(null) ");
	       } else {
		  lptr = (long int*)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++) {
		     if (lptr[k] != NULLINT)
			fprintf(outfp,"%ld ",lptr[k]);
		     else
			fprintf(outfp,"(null) ");
		  }
		  if (lptr) free(lptr);
	       }
	       break;
	    case 'S':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&ival,&n);
		  if (ival != NULLSHORT)
		     fprintf(outfp,"%d ",ival);
		  else
		     fprintf(outfp,"(null) ");
	       } else {
		  iptr = (short int *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++) {
		     if (iptr[k] != NULLSHORT)
			fprintf(outfp,"%d ",iptr[k]);
		     else
			fprintf(outfp,"(null) ");
		  }
		  if (iptr) free(iptr);
	       }
	       break;
	    case 'F':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&fval,&n);
		  if (!is_vpf_null_float(fval))
		     fprintf(outfp,"%f ",fval);
		  else
		     fprintf(outfp,"(null) ");
	       } else {
		  fptr = (float *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++) {
		     if (!is_vpf_null_float(fptr[k]))
			fprintf(outfp,"%f ",fptr[k]);
		     else
			fprintf(outfp,"(null) ");
		  }
		  if (fptr) free(fptr);
	       }
	       break;
	    case 'R':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&rval,&n);
		  if (!is_vpf_null_double(rval))
		     fprintf(outfp,"%f ",rval);
		  else
		     fprintf(outfp,"(null) ");
	       } else {
		  rptr = (double *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++) {
		     if (!is_vpf_null_double(rptr[k]))
			fprintf(outfp,"%f ",rptr[k]);
		     else
			fprintf(outfp,"(null) ");
		  }
		  if (rptr) free(rptr);
	       }
	       break;
	    case 'C':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&cval,&n);
		  fprintf(outfp,"(%f,%f) ",cval.x,cval.y);
	       } else {
		  cptr = (coordinate_type *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++)
		     fprintf(outfp,"(%f,%f) ",cptr[k].x,cptr[k].y);
		  if (cptr) free(cptr);
	       }
	       break;
	    case 'B':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&bval,&n);
		  fprintf(outfp,"(%f,%f) ",bval.x,bval.y);
	       } else {
		  bptr = (double_coordinate_type *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++)
		     fprintf(outfp,"(%f,%f) ",bptr[k].x,bptr[k].y);
		  if (bptr) free(bptr);
	       }
	       break;
	    case 'Z':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&zval,&n);
		  fprintf(outfp,"(%f,%f,%f) ",zval.x,zval.y,zval.z);
	       } else {
		  zptr = (tri_coordinate_type *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++)
		     fprintf(outfp,"(%f,%f,%f) ",
                             zptr[k].x,zptr[k].y,zptr[k].z);
		  if (zptr) free(zptr);
	       }
	       break;
	    case 'Y':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&yval,&n);
		  fprintf(outfp,"(%f,%f,%f) ",yval.x,yval.y,yval.z);
	       } else {
		  yptr = (double_tri_coordinate_type *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++)
		     fprintf(outfp,"(%f,%f,%f) ",
                             yptr[k].x,yptr[k].y,yptr[k].z);
		  if (yptr) free(yptr);
	       }
	       break;
	    case 'K':
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&kval,&n);
		  fprintf(outfp,"(%ld,%ld,%ld) ",
			  kval.id,kval.tile,kval.exid);
	       } else {
		  kptr = (id_triplet_type *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++)
		     fprintf(outfp,"(%ld,%ld,%ld)  ",
			     kptr[k].id,kptr[k].tile,kptr[k].exid);
		  free(kptr);
	       }
	       break;
	    case 'D':   /* Date */
	       if (table.header[j].count==1) {
		  get_table_element(j,row,table,&dval,&n);
		  format_date(dval,date);
		  fprintf(outfp,"%s ",date);
	       } else {
		  dptr = (date_type *)get_table_element(j,row,table,NULL,&n);
		  for (k=0;k<n;k++) {
		     format_date(dptr[k],date);
		     fprintf(outfp,"%s ",date);
		  }
		  free(dptr);
	       }
	       break;
            case 'X':
               break;
	 }
      }
      fprintf(outfp,"\n");
      free_row( row, table );
   }

   vpf_close_table( &table );
   return 0;
}



